signature PARSE =
sig

  val parse : (Lexer.token * char list) Stream.stream -> Syntax.exp Stream.stream

  exception Error of string
end;  (* signature PARSE *)

structure Parse :> PARSE = 
struct

    type exp = Syntax.exp
    
    structure S = Stream
    open Syntax
    open Lexer

    exception Error of string

    (* 
       The concrete grammar parsed by this module is as follows:

       program ::= exp0 SEMICOLON

       typ ::= NAT | LPAREN typ ARR typ RPAREN
       
       exp0    ::= exp0 APP expinf | expinf
       expinf  ::= LPAREN exp0 RPAREN
                |  NUMBER(k)
                |  VAR(s)
                |  LET VAR(s) BE exp0 IN exp0 END
                |  ZERO
                |  SUCC exp0 
                |  REC exp0 LCURLY ZERO GOESTO exp0 BAR SUCC VAR(s) WITH VAR(s) GOESTO exp0 RCURLY
                |  LAM var(S) COLON typ DOT exp0

    *)

    fun Num 0 = Zero
      | Num n = Succ ( (Num (n-1)))
       
    (* next s = (x,s'), where x is the head of s, s' the tail of s
       raises Error if stream is empty
    *)
    fun next s =
        case S.force s
          of S.Nil => raise Error "Unexpected end of stream"
           | S.Cons result => result

    fun expected s msg =
        let val ((_, l), _) = next s
        in raise Error ("Expected " ^ msg ^ " at `" ^
                        concat (map Char.toString l) ^ "'") end

    fun require s token =
            let val ((t,l), s') = next s
            in if t = token then s'
               else expected s (Lexer.tokenToString token ^ " token")
            end

    fun program s =
            let val (e, s) = exp0 ((fn x => raise Error ("Free variable `" ^ x ^ "'")), s)
                val s = require s SEMICOLON
            in (e, s) end

    and typ s = 
        case next s of
            ((NAT,_),s) => (Syntax.Nat,s)
          | ((LPAREN,_),s) => let val (t1,s) = typ s
                                  val s = require s ARR
                                  val (t2,s) = typ s
                                  val s = require s RPAREN
                              in
                                  (Syntax.Arrow(t1,t2),s)
                              end
          | ((t,_),s) => raise Error ("expected type; found " ^ (tokenToString t))

    and exp0 (g, s) =
            let val (e, s) = expinf (g, s)
            in exp0' (e, g, s) end
    and exp0' (e, g, s) =
            case next s
              of ((APP,_), s) => 
                     let val (e', s) = expinf (g, s)
                     in exp0' ( (App(e, e')), g, s) end
               | _ => (e, s)

    and expinf (g, s) = case next s
         of ((LPAREN,_), s) => let val (e, s) = exp0 (g, s)
                               val s = require s RPAREN
                           in (e, s) end
          | ((NUMBER(k),_), s) => ( (Num(k)), s)
          | ((VAR(x),_), s) => ( (Var(g x)), s)
          | ((LET,_), s) => let val (x, s) = var s
                                val s = require s BE
                                val (e1, s) = exp0 (g, s)
                                val s = require s IN
				val g' = extend(g, x)
                                val (e2, s) = exp0 (g', s)
                                val s = require s END
                            in ( (Let(e1, (g' x, e2))), s) end
          | ((ZERO,_) , s) => ( Zero , s)
          | ((SUCC,_) , s) => let val (e,s) = exp0 (g , s) in ( (Succ e) , s) end
          | ((REC,_),s) => let val (e,s) = exp0 (g,s)
                               val s = require s LEFTCURLY
                               val s = require s ZERO
                               val s = require s GOESTO
                               val (e0,s) = exp0 (g,s)
                               val s = require s BAR
                               val s = require s SUCC
                               val (x,s) = var s
                               val s = require s WITH
                               val (y,s) = var s
                               val s = require s GOESTO
                               val g' = extend(extend(g,x),y)
                               val (e1,s) = exp0 (g',s)
                               val s = require s RIGHTCURLY
                           in 
                               ( (Rec(e0,(g' x, g' y, e1),e)), s)
                           end
          | ((LAM,_),s) => let val (x, s) = var s
                               val s = require s COLON
                               val (t,s) = typ s
                               val s = require s DOT
                               val g' = extend (g,x)
                               val (e,s) = exp0 (g',s)
                           in
                               ( (Lam(t,(g' x,e))),s)
                           end
          | _ => expected s "an expression"

    and extend (g, x) = let val v = Syntax.freshNameString x in (fn y => if x = y then v else g y) end

    and var s = case next s
         of ((VAR(x),_), s) => (x, s)
          | _ => expected s "a variable name"

    (* Exported parsing function *)
          
    fun parse s = S.delay (fn () =>
            case S.force s
              of S.Nil => S.Nil
               | S.Cons _ => let val (e, s) = program s
                             in S.Cons(e, parse s) end)

end;  (* functor ParseFun *)
